# -*- coding: UTF-8 -*-
"""
Setup KD and SL for Sites and elements

USAGE
    
    python -m thyrsis.element_setup [-h]

OPTION

    -h, --help
        display this page
"""
from builtins import str

import sip
assert(sip.getapi('QVariant') == 2)
assert(sip.getapi('QString') == 2)


from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from qgis.PyQt import uic

import os
from ..database import sqlite
from ..settings import Settings

class ElementSetupWidget(QWidget):
    """Dialog to define properties of chemicals element"""

    def __init__(self, parent=None):
        """Constructor

        :param msg: parent widget to set the dialog
        :type parent: QWidget
        """
        super(ElementSetupWidget, self).__init__(parent)
        uic.loadUi(os.path.join(os.path.dirname(__file__), 'element_setup.ui'), self)

        self.buttonBox.accepted.connect(self.__accept)
        self.buttonBox.rejected.connect(self.close)

        site_db = str(Settings().value("General", "defaultSite"))
        self.__conn = sqlite.connect(site_db, "ElementSetupWidget:init")
        cur = self.__conn.cursor()
        cur.execute("SELECT nom, id FROM sites")
        self.siteComboBox.addItem('', '')
        for site, id_ in cur.fetchall():
            self.siteComboBox.addItem(site, id_)

        cur.execute("SELECT nom, id FROM elements_chimiques ORDER BY nom")
        self.chemicalComboBox.addItem('')
        for nom, id_ in cur.fetchall():
            self.chemicalComboBox.addItem(nom, id_)

        self.siteComboBox.currentIndexChanged.connect(self.__siteChanged)
        self.chemicalComboBox.currentIndexChanged.connect(self.__chemicalChanged)
    
        self.__sid = None
        self.__eid = None

    def __siteChanged(self, idx):
        """ Triggered when the site changes, actualize a combo box
        
        :param idx: idx
        :type idx: int

        :return: None
        :rtype: None
        """
        self.__sid = self.siteComboBox.itemData(idx)
        if self.__eid:
            self.__load()

    def __chemicalChanged(self, idx):
        """ Triggered when the used chemical changes
        
        :param idx: idx
        :type idx: int

        :return: None
        :rtype: None
        """
        self.__eid = self.chemicalComboBox.itemData(idx)
        if self.__sid:
            self.__load()
        
    def __load(self):
        """ Load the chemical properties
        
        :return: None
        :rtype: None
        """
        assert(self.__sid and self.__eid)
        cur = self.__conn.cursor()
        for table, minEdit, moyEdit, maxEdit in [
                ("coefficients_retention", self.kdMinEdit, self.kdMoyEdit, self.kdMaxEdit),
                ("limites_solubilite", self.slMinEdit, self.slMoyEdit, self.slMaxEdit)]:
            values = cur.execute(
                    "SELECT min, moy, max FROM {table} \
                    WHERE eid={eid} AND sid={sid}".format(
                        table= table, 
                        eid=self.__eid, 
                        sid=self.__sid)).fetchone()
            for i, edit in enumerate([minEdit, moyEdit, maxEdit]):
                edit.setText(str(values[i]) if values else "")
                edit.setStyleSheet("background-color : white;")

    def __accept(self):
        """ Save the changes on chemical data

        :return: None
        :rtype: None
        """
        if not self.__eid or not self.__sid:
            return

        cur = self.__conn.cursor()
        for table, minEdit, moyEdit, maxedit in [
                ("coefficients_retention", self.kdMinEdit, self.kdMoyEdit, self.kdMaxEdit),
                ("limites_solubilite", self.slMinEdit, self.slMoyEdit, self.slMaxEdit)]:
            # check that all required fields are ok
            missing = False
            for edit in [minEdit, moyEdit, maxedit]:
                try:
                    float(edit.text())
                    edit.setStyleSheet("background-color : white;")
                except:
                    edit.setStyleSheet("background-color : red;")
                    missing = True
            if not missing:
                exists, = cur.execute("SELECT COUNT(1) FROM {} WHERE eid={} AND sid={}".format(
                    table, self.__eid, self.__sid)).fetchone()
                cur.execute(
                        ("UPDATE {table} SET min={min_}, moy={moy_}, max={max_} WHERE eid={eid} AND sid={sid}"
                            if exists else
                         "INSERT INTO {table}(sid, eid, min, moy, max) VALUES({sid}, {eid}, {min_}, {moy_}, {max_})"
                        ).format(
                            table=table,
                            min_=minEdit.text(), 
                            moy_=moyEdit.text(), 
                            max_=maxedit.text(),
                            eid=self.__eid, 
                            sid=self.__sid)
                    )

        self.__conn.commit()

if __name__ == "__main__":
    
    import sys
    
    if len(sys.argv) == 2 and sys.argv[1] in ["-h","--help"]:
        help(sys.modules[__name__])
        exit(0)

    
    app = QApplication(sys.argv)
    
    dlg = ElementSetupWidget()
    dlg.show()
    app.exec_()

