# -*- coding: UTF-8 -*-
"""

settings thyrsis

USAGE
    python -m thyrsis.settings

"""
from builtins import str
from builtins import object

import os
import sys
import shutil
import codecs
from configparser import ConfigParser
from .log import logger

class Settings(object):
    """ Thyrsis settings"""
    def __init__(self, dbpath = None):
        """Cosntructor

        :param dbpath: db path
        :type dbpath: string
        """
        #logger.notice("settings.init with dbpath = ", dbpath)
        self.__config = ConfigParser()
        self.__config.optionxform = str
        self.__dbpath = dbpath
        self.install_dir = os.path.expanduser('~/.thyrsis')
        default_ini_file = os.path.join(self.install_dir, 'thyrsis.ini')

        if not os.path.isfile(default_ini_file):
            logger.debug("creating default settings")
            self.__default()
            self.__dbpath = None
            self.save() # create default init BEFORE assigning self.__dbpath
            self.__dbpath = dbpath

        if dbpath:
            if os.path.isfile(os.path.join(dbpath, '.thyrsis.ini')):
                logger.debug("reading", os.path.join(dbpath, '.thyrsis.ini'))
                with codecs.open(os.path.join(dbpath, '.thyrsis.ini'), encoding='utf-8') as file :
                    self.__config.read_file(file)
                logger.debug("codeHydro =", self.value("General", "codeHydro"))
                logger.debug("codeHydroCommand = ", self.value("General", "codeHydroCommand"))
            else:
                logger.debug("copying", default_ini_file, "to", os.path.join(dbpath, '.thyrsis.ini'))
                shutil.copyfile(default_ini_file, os.path.join(dbpath, '.thyrsis.ini'))
                with codecs.open(os.path.join(dbpath, '.thyrsis.ini'), encoding='utf-8') as file :
                    self.__config.read_file(file)

        else:
            logger.debug("reading", default_ini_file)
            with codecs.open(default_ini_file, encoding='utf-8') as file :
                self.__config.read_file(file)
            logger.debug("codeHydro =", self.value("General", "codeHydro"))
            logger.debug("codeHydroCommand = ", self.value("General", "codeHydroCommand"))

        self.__config.set("General", "installDir", self.install_dir)
        self.__config.set("General", "defaultSite", os.path.join(self.install_dir, 'sites.sqlite'))
        self.__config.set("General", "defaultMeshDir", os.path.join(self.install_dir, 'mesh'))

        self.__convert()
        self.save()


    def dbpath(self):
        """ Return db path

        :return: db path
        :rtype: string"""
        return self.__dbpath

    def value(self, section, option, defaultValue=None):
        """ Return value of an option 

        :param section: section name
        :type section: string
        :param option: option name
        :type option: string
        :param defaultValue: defaultValue
        :type defaultValue: string

        :return: value
        :rtype: string
        """
        #logger.debug("settings.value ", section, option)
        if self.__config.has_option(section, option):
            return self.__config.get(section, option)
        elif defaultValue is not None:
            return str(defaultValue)
        else:
            raise RuntimeError("value unknown "+section+":"+option)

    def setValue(self, section, option, value):
        """ Set value of an option 

        :param section: section name
        :type section: string
        :param option: option name
        :type option: string
        :param value: defaultValue
        :type value: string
        """
        #logger.debug("settings.setValue ", section, option, value, type(value))
        if not self.__config.has_section(section):
            self.__config.add_section(section)
        self.__config.set(section, option, str(value))

    def contains(self, section, option):
        """ Check if an option exists 

        :param section: section name
        :type section: string
        :param option: option name
        :type option: string

        :return: value
        :rtype: string
        """
        #logger.debug("settings.contains ", section, option)
        return self.__config.has_option(section, option)

    def options(self, section):
        """ list options of a section

        :param section: section name
        :type section: string
        """
        #logger.debug("settings.options")
        return self.__config.options(section)

    def allKeys(self):
        """ List all keys"""
        #logger.debug("settings.allKeys")
        options = []
        for section in self.__config.sections():
            options += self.__config.options(section)
        return options

    def save(self):
        """ Save settings in settings file"""
        #logger.debug("settings.save")
        #self.display()
        if not os.path.isdir(self.install_dir):
            os.mkdir(self.install_dir)
        if not os.path.isdir(os.path.join(self.install_dir, 'mesh')):
            os.mkdir(os.path.join(self.install_dir, 'mesh'))

        logger.debug("codeHydro =", self.value("General", "codeHydro"))
        logger.debug("codeHydroCommand = ", self.value("General", "codeHydroCommand"))
        if self.__dbpath and os.path.isdir(self.__dbpath):
            logger.debug("saving settings in", os.path.join(self.__dbpath, '.thyrsis.ini'))
            with open(os.path.join(self.__dbpath, '.thyrsis.ini'), 'w') as configfile:
                self.__config.write(configfile)
        # on sauve toujours dans ~/.thyrsis/thyrsis.ini
        logger.debug("saving settings in", os.path.join(self.install_dir, 'thyrsis.ini'))
        with open(os.path.join(self.install_dir, 'thyrsis.ini'), 'w') as configfile:
            self.__config.write(configfile)

    def __convert(self):
        #logger.debug("settings.convert")
        for section in self.__config.sections():
            for (option, value) in self.__config.items(section):
                self.setValue(section, option, value)

    def display(self):
        """ Display all settings"""
        #logger.debug("settings.display")
        for section in self.__config.sections():
            for (option, value) in self.__config.items(section):
                print(section, option, value, type(value))

    def __default(self):
        logger.notice("settings.default")
        self.__config.add_section("General")
        self.__config.set("General", "installDir", self.install_dir)
        self.__config.set("General", "defaultSite", os.path.join(self.install_dir, 'sites.sqlite'))
        self.__config.set("General", "defaultMeshDir", os.path.join(self.install_dir, 'mesh'))
        self.__config.set("General", "codeHydro", "")
        self.__config.set("General", "codeHydroCommand", "")
        self.__config.set("General", "maps", "")
        self.__config.set("General", "mpiConfig", "")
        self.__config.set("General", "nbMpiProcess", "1")
        self.__config.set("General", "nbColumnsMax", "200")
        self.__config.set("General", "console", "0")
        self.__config.set("General", "debug", "0")

        self.__config.add_section("Variables")
        self.__config.set("Variables", "potentielZSCheck", "2")
        self.__config.set("Variables", "darcyZNSCheck", "0")
        self.__config.set("Variables", "darcyZSCheck", "0")
        self.__config.set("Variables", "concentrationZNSCheck", "2")
        self.__config.set("Variables", "concentrationZSCheck", "2")
        self.__config.set("Variables", "activityZNSCheck", "0")
        self.__config.set("Variables", "saturationZNSCheck", "0")
        self.__config.set("Variables", "debitZSCheck", "1")
        self.__config.set("Variables", "potentielUnit", "m")
        self.__config.set("Variables", "darcyUnit", "m/s")
        self.__config.set("Variables", "concentrationUnit", "mg/L")
        self.__config.set("Variables", "activityUnit", "Bq/kg")
        self.__config.set("Variables", "massUnit", "kg")
        self.__config.set("Variables", "concentrationScale", "1e-3")
        self.__config.set("Variables", "darcyScale", "1e-3")
        self.__config.set("Variables", "concentrationClasses", "10")
        self.__config.set("Variables", "potentielClasses", "10")
        self.__config.set("Variables", "darcyClasses", "10")
        self.__config.set("Variables", "maskUnitsCheck", "0")

        self.__config.add_section("Animation")
        self.__config.set("Animation", "framePerSecond", "25")

        self.__config.add_section("Matplotlib")
        self.__config.set("Matplotlib", "ticksLabelSize", "10")
        self.__config.set("Matplotlib", "ylabelFontSize", "10")
        self.__config.set("Matplotlib", "legendFontSize", "10")

if __name__ == "__main__":

    import sys

    settings = Settings()

