# coding=utf-8
"""
Crée des colonnes d'injection regroupant le domaine par épaisseur et perméabilité

USAGE

    python -m thyrsis.create_injection_zones database.sqlite
"""
from builtins import str
from builtins import range

from ..database import sqlite
from ..database import create_injection_triggers, delete_injection_triggers
from ..utilities import Timer
from ..log import logger
import sys
import os
import getopt
import numpy
from collections import defaultdict
from shapely.geometry import LineString, MultiLineString, MultiPolygon, Point
from shapely.ops import linemerge, polygonize

def create_injection_zones(dbname, debug=False):
    """Launch parrallel process as configured

    :param dbname: path to database
    :type dbname: string
    :param debug: verbose flag
    :type debug: bool
    """
    from rtree import index

    logger.notice(u"\nCREATION DES ZONES D'INJECTION")

    conn = sqlite.connect(dbname)
    cur = conn.cursor()

    project_SRID = str(cur.execute("SELECT DISTINCT st_srid(GEOMETRY) FROM noeuds").fetchone()[0])

    # get mesh vertices
    cur.execute("SELECT X(GEOMETRY), Y(GEOMETRY) FROM noeuds ORDER BY OGC_FID")
    vtx = numpy.require(cur.fetchall(), numpy.float32, 'F')

    # get triangulation of mesh elements
    cur.execute("SELECT a, b, c FROM mailles WHERE d IS NULL ORDER BY OGC_FID")
    idx1 = numpy.require(cur.fetchall(), numpy.int32, 'F')
    cur.execute("SELECT a, b, d FROM mailles WHERE d IS NOT NULL ORDER BY OGC_FID")
    idx2 = numpy.require(cur.fetchall(), numpy.int32, 'F')
    cur.execute("SELECT a, d, c FROM mailles WHERE d IS NOT NULL ORDER BY OGC_FID")
    idx3 = numpy.require(cur.fetchall(), numpy.int32, 'F')
    triangles = None
    if len(idx1) and len(idx2): # triangles and quads
        triangles = numpy.concatenate((idx1, idx2, idx3))
    elif len(idx1): # only triangles
        triangles = idx1
    else: # only quads
        triangles = numpy.concatenate((idx2, idx3))
    triangles -= 1

    timer = Timer()
    cur.execute("DELETE FROM noeuds_injections")
    cur.execute("DELETE FROM injections")
     
    cur.execute("SELECT pas_echant_zns, rapport_max_permeabilite, permeabilite_zns_uniforme FROM simulations")
    pas_echant_zns, rapport_max_permeabilite, permeabilite_zns_uniforme = cur.fetchone()
    if pas_echant_zns <= 0:
        groups = numpy.array(list(range(len(vtx))))
        nb_group = len(vtx)
    else:
        logger.debug("pas_echant_zns, rapport_max_permeabilite, permeabilite_zns_uniforme = ",
            pas_echant_zns, rapport_max_permeabilite, permeabilite_zns_uniforme)
        # group by epaisseur_zns first and permeabilities second
        groups = numpy.array([-1]*len(vtx))
        cur.execute("""
            SELECT 
                OGC_FID,
                MAX(0, epaisseur_zns)
            FROM noeuds
            ORDER BY epaisseur_zns ASC""")
        nb_group = 0
        current_epaisseur_zns = 0
        group_nodes = defaultdict(list)
        for ogc_fid, epaisseur_zns in cur.fetchall():
            if epaisseur_zns > current_epaisseur_zns + pas_echant_zns:
                logger.debug("gpr {} epaisseur_zns {}".format(nb_group, epaisseur_zns))
                nb_group += 1
                current_epaisseur_zns = epaisseur_zns
            groups[ogc_fid-1] = nb_group
            group_nodes[nb_group].append(ogc_fid-1)

        logger.notice(timer.reset("created {} groups {}".format(len(list(group_nodes.keys())), nb_group)))
        cur.execute("""
            SELECT 
                Coalesce(permeabilite_zns, permeabilite_x) AS permeabilite 
            FROM noeuds
            ORDER BY OGC_FID ASC""")
        perm = numpy.array(cur.fetchall()).ravel()
        # split group if permeabilite is > rapport_max_permeabilite
        for g in range(nb_group+1):
            grp_nodes = numpy.array(group_nodes[g])
            perm_ = perm[grp_nodes]
            idx = numpy.argsort(perm_)
            logger.debug("gpr {} min perm {} max perm {}".format(g, perm_[idx[0]], perm_[idx[-1]]))
            if perm_[idx[0]]*rapport_max_permeabilite < perm_[idx[-1]]:
                # split group
                start_idx = idx[0]
                for n in idx:
                    if perm_[start_idx]*rapport_max_permeabilite < perm_[n]:
                        start_idx = n
                        nb_group += 1
                        logger.debug("   splittin gpr {} at perm {} -> {}".format(g, perm_[start_idx], nb_group))
                    if start_idx != idx[0]:
                        groups[grp_nodes[n]] = nb_group

        nb_group += 1

        insert = []
        for i, g in enumerate(groups):
            insert.append((int(i+1), int(g+1)))
        cur.executemany("INSERT INTO noeuds_injections(noeud, injection) VALUES(?, ?)", insert)

    cur.executemany("UPDATE noeuds SET groupe=? WHERE OGC_FID=?",  [(g, i+1) for i, g in enumerate(groups)]) 

    logger.notice(timer.reset("created %d groups"%(nb_group)))
    assert numpy.max(groups) + 1 == nb_group
    assert numpy.min(groups) == 0
    count = 0
    cut = defaultdict(list)
    edges = defaultdict(list)
    lines = []
    triple_points = []
    key_points = []
    for i, triangle in enumerate(triangles):
        sys.stdout.write("\r%3d%%"%(int(float(i)/len(triangles)*100)))
        line = []
        for e in [tuple(sorted((triangle[0], triangle[1]))),
                  tuple(sorted((triangle[1], triangle[2]))),
                  tuple(sorted((triangle[2], triangle[0])))]:
           edges[e].append(i)
           if groups[e[0]] != groups[e[1]]:
               cut[e].append(i)
               line.append(.5*(vtx[e[0]]+vtx[e[1]]))
               key_points += [e[0], e[1]]
        if len(line):
            assert len(line) in (2,3)
            if len(line) == 3:
                center = (vtx[triangle[0]]+vtx[triangle[1]]+vtx[triangle[2]])/3
                triple_points.append(center)
                for j in range(3):
                    lines.append([center, .5*(vtx[triangle[j]]+vtx[triangle[(j+1)%3]])])
            else:
                lines.append(line)


    for e, tri in edges.items():
        if len(tri) == 1:
            if e in cut:
                lines.append([vtx[e[0]], .5*(vtx[e[0]]+vtx[e[1]])]) 
                lines.append([.5*(vtx[e[0]]+vtx[e[1]]), vtx[e[1]]])
                pass
            else:
                lines.append([vtx[e[0]], vtx[e[1]]])
               
    border = [edge[0] for edge, tri in edges.items() if len(tri) == 1]
    logger.debug(timer.reset("%d cut elements, %d border %d total"%(len(cut), len(border), len(triangles))))

    if debug:
        import matplotlib.pyplot as plt
        for line in lines:
            plt.plot([p[0] for p in line], [p[1] for p in line], 'v-')

        plt.plot([p[0] for p in triple_points], [p[1] for p in triple_points], 'o')
        plt.show()

    ml = linemerge(MultiLineString(lines))
    polygons = polygonize(ml)
    logger.debug(timer.reset("limerge & polygonize"))

    idx = index.Index() 
    for i in key_points:
        idx.insert(i, (vtx[i][0], vtx[i][1], vtx[i][0], vtx[i][1]))

    result = [[] for g in range(nb_group)]       
    for polygon in polygons:
        for j in idx.intersection(polygon.bounds):
            if polygon.intersects(Point(vtx[j])):
                result[groups[j]].append(polygon)
                break # first intersecting point
    logger.debug(timer.reset("indexing"))

    delete_injection_triggers(cur)

    cur.execute("delete from injections")

    insert = []
    for g, p in enumerate(result):
        if len(p) > 0 :
            insert.append((MultiPolygon(p).wkt, "injection_{}".format(g)))
    logger.debug(timer.reset("polygons aggregation"))
    cur.executemany("insert into injections(GEOMETRY, nom) values(GeomFromText(?, %s), ?)"%(project_SRID), insert)
    logger.debug(timer.reset("polygons insertion"))
    cur.execute("""
        UPDATE injections SET
            permeabilite = (
              SELECT Coalesce(EXP(SUM(LN(n.permeabilite_zns))/COUNT(1)),
                               EXP(SUM(LN(n.permeabilite_x))/COUNT(1)))
              FROM noeuds AS n, noeuds_injections AS ni 
              WHERE ni.injection = injections.OGC_FID
              AND n.OGC_FID = ni.noeud ),
            infiltration = (
              SELECT AVG(n.infiltration)
              FROM noeuds AS n, noeuds_injections AS ni 
              WHERE ni.injection = injections.OGC_FID
              AND n.OGC_FID = ni.noeud ),
            epaisseur = (
              SELECT MIN(n.epaisseur_zns)
              FROM noeuds AS n, noeuds_injections AS ni 
              WHERE ni.injection = injections.OGC_FID
              AND n.OGC_FID = ni.noeud ),
            altitude = (
              SELECT AVG(n.altitude)
              FROM noeuds AS n, noeuds_injections AS ni
              WHERE ni.injection = injections.OGC_FID
              AND n.OGC_FID = ni.noeud ),
            debut = (SELECT debut FROM simulations)
        ;
        """)


    create_injection_triggers(cur)
    conn.commit()
    conn.close()

if __name__ == "__main__":
    try:
        optlist, args = getopt.getopt(sys.argv[1:],
                "hd",
                ["help", "debug"])
    except Exception as e:
        sys.stderr.write(str(e)+"\n")
        exit(1)

    logger.enable_console(True)
    logger.set_level('debug')

    optlist = dict(optlist)
    debug = "-d" in optlist or "--debug" in optlist

    if "-h" in optlist or "--help" in optlist:
        help(sys.modules[__name__])
        exit(0)

    if len(args) != 1:
        sys.stderr.write("error: missing database\n")
        exit(1)
    dbname = args[0]

    # use provider for node coord and triangle read

    create_injection_zones(dbname, debug)

