# -*- coding: utf-8 -*-
"""
Inversion : détermination du champ de perméabilité par inversion des données piézométriques

USAGE:
    python -m thyrsis.simulation.hynverse database.sqlite [-t, -c, -v]

OPTIONS

   -t : test
   -r : run inversion
   -v : verbose
"""
from builtins import str
from builtins import range
from builtins import object

SRID = '27572' # LAMBERT II extended
PROCESS = None

import os
import sys
import string
import numpy
import shutil
import subprocess
from math import log, sqrt, exp as exp_

from ..database import sqlite
from ..simpleprovider import SimpleProvider

from ..utilities import Timer, parse_file
from ..database import sqlite
from ..log import logger
from .metis import Metis
from .openfoam import OpenFoam
from ..settings import Settings

import atexit

LOCK_NAME="hynverse"

@atexit.register
def terminate_process():
    """Triggered if calcul is aborted"""
    if PROCESS is not None and PROCESS.poll() is None:
        logger.warning("metis process aborted")
        PROCESS.terminate()

class Zone_perm(object):
    """ Class object for permeability zone"""
    def __init__(self, num, perm, pmin, pmax, liste_pp):
        """Constructor

        :param num: zone id
        :type num: int
        :param perm: permeability
        :type perm: float
        :param pmin: min permeability tolerated
        :type pmin: float
        :param pmax: max permeability tolerated
        :type pmax: float
        :param liste_pp: pilot point list
        :type liste_pp: list
        """
        self.num = num
        self.perm = perm
        self.pmin = pmin
        self.pmax = pmax
        self.liste_pp = liste_pp

class Point_pilote(object):
    """ Class object for pilot point"""
    def __init__(self, x, y, num, nom, groupe, zone, hmes, hcal=0):
        """Constructor

        :param x: x coord
        :type x: float
        :param y: y coord
        :type y: float
        :param num: zone id
        :type num: int
        :param nom: pilot point name
        :type nom: string
        :param groupe: pilot point group name
        :type groupe: string
        :param zone: zone id
        :type zone: int
        :param hmes: piezometric altitude measured
        :type hmes: float
        """
        self.x = x
        self.y = y
        self.num = num  # numero dans liste_pp totale
        self.nom = nom
        self.groupe = groupe
        self.zone = zone
        self.hmes = hmes
        self.hcal = hcal if hcal else None

def distance(x1, y1, x2, y2):
    """ Compute distance between two points

    :param x1: x1 coord
    :type x1: float
    :param y1: y1 coord
    :type y1: float
    :param x2: x2 coord
    :type x2: float
    :param y2: y2 coord
    :type y2: float

    :return: distance
    :rtype: float
    """
    return sqrt((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1))

def distance2(x1, y1, x2, y2):
    """ Compute distance² between two points

    :param x1: x1 coord
    :type x1: float
    :param y1: y1 coord
    :type y1: float
    :param x2: x2 coord
    :type x2: float
    :param y2: y2 coord
    :type y2: float

    :return: distance²
    :rtype: float
    """
    return (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1)

def poids(x, y, nearest, xy):
    """ Compute weight between point, used for indexing

    :param x: x1 coord
    :type x: float
    :param y: y1 coord
    :type y: float
    :param nearest: nearest points list
    :type nearest: list
    :param xy: pilots point xy
    :type xy: list

    :return: weight
    :rtype: float
    """
    if distance2(x, y, xy[nearest[0]][0], xy[nearest[0]][1]) == 0:
        return numpy.array([1]+[0 for i in range(len(nearest)-1)])
    else:
        pds = numpy.array([1./distance2(x, y, xy[n][0], xy[n][1]) for n in nearest])
        return pds/numpy.sum(pds)

def fun1(x,y):
    """ Return coefficient 1/-1

    :param x: x1 coord
    :type x: float
    :param y: y1 coord
    :type y: float

    :return: coefficient
    :rtype: int
    """
    if x > y:
        return 1.
    elif x < -y:
        return -1
    else:
        return 0.

class Hynverse(object):
    """ Process class for inversion calculation"""
    def __init__(self, code, perm):
        """Constructor

        :param code: Simulation code
        :type code: Metis or OpenFoam
        :param perm: permeability field provided flag
        :type perm: bool

        """
        logger.notice("\nINITIALISATION")
        self.init_flag = True
        self.__code = code
        self.__database = code.database
        self.__parametres = self.hynverse_param()
        self.perm = perm
        if self.perm :
            self.__parametres['niter']=1
            logger.notice('Using a permeability field to create the model')
        logger.notice(u'parametres = ', self.__parametres)
        self.__tmpdir = self.hynverse_tmpdir()
        if not self.perm :
            if not self.init_flag :
                return
            self.__idx_n, self.__xy_n = self.index_nodes()
            self.__idx_m, self.__xy_m = self.index_mailles()
            self.__idx_pp, self.__xy_pp, self.__liste_pp = self.hynverse_pp()
            if not self.init_flag:
                return
            self.__idx_pppp = self.index_pppp()
            self.__idx_npp, self.__pds_npp = self.index_npp()
            self.__idx_ppm, self.__pds_ppm = self.index_ppm()
            self.__idx_ppn, self.__pds_ppn = self.index_ppn() # pour le calcul de la permeabilite aux noeuds
            self.__dict_zones = self.hynverse_zones()
        code.create_hynverse_files(self.__tmpdir, self.__parametres)
        if self.__parametres['niter'] <= 0:
            self.init_flag = False

        #with ExclusiveLock(LOCK_NAME):
        with sqlite.connect(self.__database) as conn:
            conn.cursor().execute("DELETE from hynverse_erreurs")
            conn.commit()

    def hynverse_param(self):
        """ Return hynverse parameters from database

        :return: parameters
        :rtype: dict
        """
        # parametres de l'inversion
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            names = cur.execute("PRAGMA table_info(hynverse_parametres)").fetchall()
            param_names = [names[i][1] for i in range(1, len(names))]
            logger.notice('param_names = ', param_names)
            param_values = cur.execute("SELECT %s FROM hynverse_parametres ORDER BY OGC_FID"%(','.join(param_names))).fetchone()
            return {param_names[i]: param_values[i] for i in range(len(param_names))}

    def get_params(self):
        """ Return hynverse parameters

        :return: parameters
        :rtype: dict
        """
        return self.__parametres

    def hynverse_tmpdir(self):
        """ Create hynverse temp dir

        :return: tempdir
        :rtype: string
        """
        # création du répertoire de calcul
        dbname = self.__database
        dir_name = os.path.dirname(os.path.abspath(dbname))
        database = os.path.join(dir_name, dbname)
        tmpdir = os.path.join(dir_name, os.path.basename(dbname)[:-7]+'_tmp')
        if os.path.isdir(tmpdir):
            shutil.rmtree(tmpdir)
        os.makedirs(tmpdir)
        logger.debug("tmpdir = ", tmpdir)

        # fichiers utilisateur
        for filename in os.listdir(dir_name):
            if filename[0:2] == 'u_':
                if os.path.isdir(filename):
                    shutil.copytree(os.path.join(dir_name, filename), os.path.join(tmpdir, filename[2:]))
                else:
                    logger.notice(u'Copie du fichier ', filename)
                    shutil.copyfile(os.path.join(dir_name, filename), os.path.join(tmpdir, filename[2:]))

        # infiltration
        #with ExclusiveLock(LOCK_NAME):
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            if self.__parametres['infiltration'] > 0.:
                logger.notice(u'infiltration homogene')
                cur.execute("UPDATE noeuds SET infiltration = %g"%(self.__parametres['infiltration']))
                cur.execute("UPDATE mailles SET infiltration = %g"%(self.__parametres['infiltration']))
            elif os.path.exists(os.path.join(tmpdir, 'infiltration.node.sat')):
                logger.notice(u'lecture du fichier infiltration.node.sat')
                nnoeuds = code.cur.execute("SELECT COUNT(1) FROM noeuds").fetchone()[0]
                infiltration = parse_file(os.path.join(tmpdir, 'infiltration.node.sat'), 1, nnoeuds)
                cur.executemany("UPDATE noeuds SET infiltration = ? WHERE OGC_FID = ? ", infiltration)
                node_to_elem(cur, infiltration)
                os.remove(os.path.join(tmpdir, 'infiltration.node.sat'))
            elif os.path.exists(os.path.join(tmpdir, 'infiltration.elem.sat')):
                logger.notice(u'lecture du fichier infiltration.elem.sat')
                nmailles = code.cur.execute("SELECT COUNT(1) FROM mailles").fetchone()[0]
                infiltration = parse_file(os.path.join(tmpdir, 'infiltration.elem.sat'), 1, nmailles)
                cur.executemany("UPDATE mailles SET infiltration = ? WHERE OGC_FID = ? ", infiltration)
                elem_to_node(cur, infiltration)
                os.remove(os.path.join(tmpdir, 'infiltration.elem.sat'))
            else:
                logger.error("infiltration nulle et fichier u_infiltration.node.sat ou u_infiltration.elem.sat absent")
                self.init_flag = False

            # epaisseur aquifere
            cur.execute("UPDATE noeuds SET epaisseur_aquifere = altitude - altitude_mur")
            cur.execute("UPDATE mailles SET epaisseur_aquifere = altitude - altitude_mur")

            conn.commit()
        return tmpdir

    def index_nodes(self):
        """ Compute index from nodes coordinates

        :return idx_n: nodes index
        :rtype idx_n: rtree.Index
        :return xy_n: nodes coordinates list
        :rtype xy_n: list
        """
        # indexage des noeuds
        from rtree import index
        idx_n = index.Index()
        i_n = 0
        xy_n = []
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            cur.execute("SELECT X(GEOMETRY), Y(GEOMETRY) FROM noeuds ORDER BY OGC_FID")
            for x, y in cur.fetchall():
                idx_n.insert(i_n, (x, y, x, y))
                xy_n.append((x,y))
                i_n += 1
        return idx_n, xy_n

    def index_mailles(self):
        """ Compute index from nodes coordinates

        :return idx_m: mesh index
        :rtype idx_m: rtree.Index
        :return xy_m: mesh coordinates list
        :rtype xy_m: list
        """
        # indexage des centres de maille
        from rtree import index
        idx_m = index.Index()
        i_m = 0
        xy_m = []
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            cur.execute("SELECT X(Centroid(GEOMETRY)), Y(Centroid(GEOMETRY)) FROM mailles ORDER BY OGC_FID")
            for x, y in cur.fetchall():
                idx_m.insert(i_m, (x, y, x, y))
                xy_m.append((x,y))
                i_m += 1
        return idx_m, xy_m

    def hynverse_pp(self):
        """ Compute index from pilot point

        :return idx_pp: pp index
        :rtype idx_pp: rtree.Index
        :return xy_pp: pp coordinates list
        :rtype xy_pp: list
        :return list_pp: Point_pilote list
        :rtype list_pp: list
        """
        # points pilotes
        from rtree import index

        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            # deleting pilote points from an precedent potentiel_reference
            cur.execute("DELETE FROM points_pilote WHERE SUBSTR(nom,1,6)='SUPPL_'")

            liste_pp = []
            idx_pp = index.Index()
            n_pp = 0
            xy_pp = []
            n_pp_groupes = 0
            n_pp_zones = 0
            cur.execute("""SELECT X(GEOMETRY), Y(GEOMETRY), nom, groupe, zone,
                    altitude_piezo_mesuree FROM points_pilote ORDER BY OGC_FID""")
            for x, y, nom, groupe, zone, hmes in cur.fetchall():
                liste_pp.append(Point_pilote(x, y, n_pp+1, nom, groupe, zone, hmes))
                idx_pp.insert(n_pp, (x, y, x, y))
                xy_pp.append((x,y))
                n_pp_groupes = max(n_pp_groupes, groupe)
                n_pp_zones = max(n_pp_zones, zone)
                n_pp += 1
            n_pp_ref = n_pp
            logger.notice(u'Nombre de points pilote = ', n_pp)

            # potentiel de reference
            dmesh = self.__parametres['d_mesh']
            dmesh2 = dmesh*dmesh
            n_pp_sup = 0
            if cur.execute("SELECT COUNT(*) from noeuds WHERE potentiel_reference is NULL").fetchone()[0] == 0:
                logger.notice(u'Ajout de points pilotes')
                # indexage des noeuds à potentiel imposé
                idx_pot_imp = index.Index()
                nodes_pot_imp = list(set([x[0]-1 for x in \
                    cur.execute("SELECT nid from potentiel_impose order by nid").fetchall()]))
                for i,node in enumerate(nodes_pot_imp):
                    idx_pot_imp.insert(i, (self.__xy_n[node][0], self.__xy_n[node][1], self.__xy_n[node][0], self.__xy_n[node][1]))

                provider = SimpleProvider('dbname=%s crs=epsg:27572 column=%s'
                                %(self.__database,'potentiel_reference'))

                #with ExclusiveLock(LOCK_NAME):
                for m,xy in enumerate(self.__xy_m):
                    # on exclut les mailles trop proches des noeuds à potentiel imposé
                    nearest = list(idx_pot_imp.nearest((xy[0], xy[1], xy[0], xy[1]), 1))
                    if distance2(xy[0], xy[1], self.__xy_n[nearest[0]][0], self.__xy_n[nearest[0]][1]) < dmesh2:
                        continue
                    nearest = list(idx_pp.nearest((xy[0], xy[1], xy[0], xy[1]), 1))
                    if distance2(xy[0], xy[1], liste_pp[nearest[0]].x, liste_pp[nearest[0]].y) > dmesh2:
                        if n_pp == n_pp_ref:
                            n_pp_groupes += 1
                        n_pp += 1
                        n_pp_sup += 1
                        n_pp_zones += 1
                        value = float(provider.valuesAt(xy))
                        nom = 'SUPPL_'+str(n_pp_sup)
                        print("new pilot point = ", xy[0], xy[1], nom, n_pp_groupes, n_pp_zones, value)
                        liste_pp.append(Point_pilote(xy[0], xy[1], n_pp, nom, n_pp_groupes, n_pp_zones, value))
                        idx_pp.insert(n_pp-1, (xy[0], xy[1], xy[0], xy[1]))
                        xy_pp.append((xy[0], xy[1]))
                        cur.execute("""INSERT INTO points_pilote
                            (GEOMETRY, nom, groupe, zone, altitude_piezo_mesuree)
                            VALUES ( MakePoint( ?, ?, {}), ?, ?, ?, ? )""".format(project_SRID), \
                            (xy[0], xy[1], nom, n_pp_groupes, n_pp_zones, value))

                conn.commit()

        logger.notice(u'Nombre de points pilote ajoutés = ', n_pp_sup)
        logger.notice(u'Nombre total de points pilote = ', n_pp)
        if n_pp == 0:
            logger.notice(u'Aucun point pilote !')
            self.init_flag = False

        return idx_pp, xy_pp, liste_pp

    def index_pppp(self):
        """ Compute index of pilot point from themself

        :return idx_pppp: pp index
        :rtype idx_pppp: list
        """
        # index des points pilotes par rapport à eux-mêmes
        # et nombre de plus proches dans un rayon donné (dv_pp)
        nnearest = 100
        dv_pp = self.__parametres['dv_pp']
        dv_pp2 = dv_pp*dv_pp
        idx_pppp = []
        for pp in self.__liste_pp:
            nearest = list(self.__idx_pp.nearest((pp.x, pp.y, pp.x, pp.y), nnearest))
            idx_pppp.append([nearest[0]])
            for n in nearest[1:]:
                if distance2(pp.x, pp.y, self.__xy_pp[n][0], self.__xy_pp[n][1]) <= dv_pp2:
                    idx_pppp[-1].append(n)
        return idx_pppp

    def index_ppm(self):
        """ Compute index of pilot point from center mesh element

        :return idx_ppm: ppm index
        :rtype idx_ppm: list
        """
        # index des points pilote les plus proches des centres de mailles
        logger.notice(u'Indexage des points pilote les plus proches des centres de mailles')
        idx_ppm = []
        pds_ppm = []
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            cur.execute("SELECT X(Centroid(GEOMETRY)), Y(Centroid(GEOMETRY)) FROM mailles ORDER BY OGC_FID")
            for x, y in cur.fetchall():
                nearest = list(self.__idx_pp.nearest((x, y, x, y), self.__parametres['nv_ppm']))
                idx_ppm.append(nearest)
                pds_ppm.append(poids(x, y, nearest, self.__xy_pp))

        return idx_ppm, pds_ppm

    def index_ppn(self):
        """ Compute index of pilot point from nodes

        :return idx_ppn: ppn index
        :rtype idx_ppn: list
        """
        # index des points pilote les plus proches des noeuds (pour le calcul de la permeabilite aux noeuds)
        logger.notice(u'Indexage des points pilote les plus proches des noeuds')
        idx_ppn = []
        pds_ppn = []
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            cur.execute("SELECT X(Centroid(GEOMETRY)), Y(Centroid(GEOMETRY)) FROM noeuds ORDER BY OGC_FID")
            for x, y in cur.fetchall():
                nearest = list(self.__idx_pp.nearest((x, y, x, y), self.__parametres['nv_ppm']))
                idx_ppn.append(nearest)
                pds_ppn.append(poids(x, y, nearest, self.__xy_pp))

        return idx_ppn, pds_ppn

    def index_npp(self):
        """ Compute index of nearest nodes to pilot point

        :return idx_npp: npp index
        :rtype idx_npp: list
        :return pds_npp: weight of nearest nodes
        :rtype pds_npp: list

        """
        # index des noeuds les plus proches des points pilote
        logger.notice(u'Indexage des noeuds les plus proches des points pilote')
    
        idx_npp = []
        pds_npp = []
        for pp in self.__liste_pp:
            nearest = list(self.__idx_n.nearest((pp.x, pp.y, pp.x, pp.y), self.__parametres['nv_npp']))
            idx_npp.append(nearest)
            pds_npp.append(poids(pp.x, pp.y, nearest, self.__xy_n))

        return idx_npp, pds_npp

    def hynverse_zones(self):
        """ Create hynverse zones from pilot points data

        :return: zones dictionnary
        :rtype: dict
        """
        # zones
        dict_zones = dict()
        #with ExclusiveLock(LOCK_NAME):        
        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            cur.execute("DELETE from zones")
            for pp in  self.__liste_pp:
                if pp.zone in list(dict_zones.keys()):
                    dict_zones[pp.zone].liste_pp.append(pp)
                elif pp.zone > 0:
                    dict_zones[pp.zone] = Zone_perm(pp.zone, self.__parametres["permini"], \
                             self.__parametres["permin"], self.__parametres["permax"], [pp])
                    cur.execute( \
                    "INSERT INTO zones(numero, permeabilite, permeabilite_min, permeabilite_max) VALUES(?, ?, ?, ?)",
                    (pp.zone, self.__parametres["permini"], self.__parametres["permin"], self.__parametres["permax"]))

            conn.commit()
        logger.notice(u'Nombre de zones = ', len(dict_zones))

        return dict_zones

    def calcul(self):
        """ Hynverse proccess """
        sat_dir = self.__tmpdir
        if not self.perm :
            dict_zones = self.__dict_zones
            liste_pp = self.__liste_pp

            if len(dict_zones) == 0:
                logger.notice(u'Aucune zone définie')
                return

        it = 1
        ie = 1
        imin = 1
        nessais = 0
        icalc = self.__parametres['icalc']
        icalc0 = icalc
        if icalc != 1:
            icalc = 1
        alfa = self.__parametres['alfa']
        alfmin = self.__parametres['alfmin']
        errlim = self.__parametres['errlim']
        erreur = errlim + 1.
        terr = self.__parametres['terr']
        errmin = 1.e10
        nessaismax = self.__parametres['nessaismax']
        niter = self.__parametres['niter']

        if not self.perm :
            delt0 = numpy.zeros(len(dict_zones), dtype=numpy.float32)
            delt1 = numpy.zeros(len(dict_zones), dtype=numpy.float32)
            perm0 = numpy.array([dict_zones[i].perm for i in sorted(dict_zones.keys())])
            perm1 = numpy.zeros(len(dict_zones), dtype=numpy.float32)
            popt = numpy.zeros(len(dict_zones), dtype=numpy.float32)
            permin = numpy.array([dict_zones[i].pmin for i in sorted(dict_zones.keys())])
            permax = numpy.array([dict_zones[i].pmax for i in sorted(dict_zones.keys())])
      
        while it <= niter \
            and erreur >= errlim \
            and alfa >= alfmin \
            and nessais < nessaismax:

            logger.notice('ITERATION ', str(it))

            if not self.perm :
                # détermination des perméabilités en chaque zone
                deltm = numpy.max(numpy.abs(delt0))
                if icalc == 1:
                    coef = numpy.array([fun1(delt0[i-1], 1.e-3*errlim) for i in sorted(dict_zones.keys())])
                    perm1 = perm0 + alfa*coef*perm0
                elif deltm > 0.:
                    perm1 = perm0 + alfa*delt0*perm0/deltm
                perm1 = numpy.maximum(perm1, permin)
                perm1 = numpy.minimum(perm1, permax)

                for j,i in enumerate(sorted(dict_zones.keys())):
                    dict_zones[i].perm = perm1[j]

                # construction du champ de perméabilités aux mailles
            
                # lissage par moyenne mobile
                liste_ppk = [] # perméabilités aux pp
                for i,pp in enumerate(liste_pp):
                    p = 0.
                    for n in self.__idx_pppp[i]:
                        p += log(dict_zones[liste_pp[n].zone].perm)
                    liste_ppk.append(p/len(self.__idx_pppp[i]))

                # on réaffecte la perméabilité de chaque zone par moyenne sur les pp de la zone
                for z, zone in list(dict_zones.items()):
                    p = 0.
                    for pp in zone.liste_pp:
                        p += liste_ppk[pp.num - 1]
                    dict_zones[z].perm = exp_(p/len(zone.liste_pp))

                # permeabilites aux mailles
                liste_pm = []
                for j,m in enumerate(self.__idx_ppm):
                    p = 0.
                    for k,i in enumerate(m):
                        p += log(dict_zones[liste_pp[i].zone].perm)*self.__pds_ppm[j][k]
                    p = exp_(p)
                    liste_pm.append((p,p,j+1))

                #with ExclusiveLock(LOCK_NAME):
                with sqlite.connect(self.__database) as conn:
                    cur = conn.cursor()
                    #self.__code.database_connect()
                    logger.notice(u'updating permeabilities at mesh')
                    self.__code.create_permeabilities_sat(sat_dir, liste_pm)
                    cur.executemany(
                        "UPDATE mailles SET permeabilite_x=?, permeabilite_y=? WHERE OGC_FID=?", liste_pm)

                    # permeabilites aux noeuds
                    liste_pn = []
                    for j,m in enumerate(self.__idx_ppn):
                        p = 0.
                        for k,i in enumerate(m):
                            p += log(dict_zones[liste_pp[i].zone].perm)*self.__pds_ppn[j][k]
                        p = exp_(p)
                        liste_pn.append((p,p,j+1))

                    logger.notice(u'updating permeabilities at nodes')
                    cur.executemany(
                        "UPDATE noeuds SET permeabilite_x=?, permeabilite_y=? WHERE OGC_FID=?", liste_pn)
                    conn.commit()

                # calcul hydro

                self.__code.compute_hynverse(sat_dir)

                # calcul des potentiels aux noeuds

                logger.notice(u'Calcul du potentiel aux points pilote')
                potentiel = self.__code.read_potential_at_nodes(sat_dir, [0])
                #potentiel = self.__code.read_file(os.path.join(sat_dir, 'potentiel.sat'), [0])

                # calcul des potentiels aux points pilotes et de l'erreur

                erreur = 0.
                logger.debug('{0:20s} {1:5s} {2:5s} {3:8s} {4:8s} {5:8s} {6:s}'\
                    .format('nom', 'num', 'zone', 'hmes', 'hcal', 'hc-hm', 'perm'))
                for iz,zone in enumerate(sorted(dict_zones.keys())):
                    delt1[iz] = 0.
                    for pp in dict_zones[zone].liste_pp:
                        hc = 0.
                        for i,n in enumerate(self.__idx_npp[pp.num-1]):
                            hc += potentiel[0,n]*self.__pds_npp[pp.num-1][i]
                        pp.hcal = hc

                        dh = pp.hcal - pp.hmes
                        if pp.nom[:6] != 'SUPPL_':
                            logger.debug('{0:20s} {1:5d} {2:5d} {3:8.2f} {4:8.2f} {5:8.2f} {6:.2e}'\
                                .format(pp.nom, pp.num, pp.zone, pp.hmes, pp.hcal, dh, dict_zones[pp.zone].perm))
                        delt1[iz] += dh
                        if icalc0 == 2:
                            erreur += dh*dh
                        else:
                            erreur += abs(dh)

                    delt1[iz] = delt1[iz]/len(dict_zones[zone].liste_pp)

                if icalc0 == 2:
                    erreur = sqrt(erreur)
                else:
                    erreur = erreur/len(liste_pp)

                logger.notice("erreur aux points pilote = ", erreur)

                #with ExclusiveLock(LOCK_NAME):
                with sqlite.connect(self.__database) as conn:
                    cur = conn.cursor()

                    cur.execute(
                        """INSERT INTO hynverse_erreurs(iteration, iteration_vraie, nessais, erreur, alfa, terr)
                        VALUES(?,?,?,?,?,?)""", (it, ie, nessais, erreur, alfa, terr))
                    # points pilote
                    for pp in liste_pp:
                        cur.execute(
                        """UPDATE points_pilote SET altitude_piezo_calculee=?, difference_calcul_mesure=?,
                            permeabilite=? WHERE OGC_FID=?""", (pp.hcal, pp.hcal - pp.hmes, dict_zones[pp.zone].perm, pp.num))
                    conn.commit()


                if erreur < terr*errmin:
                    logger.notice("acceptation")
                    nessais = 0
                    perm0 = perm1
                    numpy.copyto(delt0, delt1)

                    if erreur < errmin:
                        imin = ie
                        errmin = erreur
                        popt = perm0

                    ie += 1

                else:
                    if icalc0 != 1 and icalc == 1:
                        logger.notice("erreur trop grande, passage en mode 3")
                        icalc = icalc0
                        alfa = 0.5
                    else:
                        logger.notice("erreur trop grande, diminution du pas de calcul")
                        alfa = 0.5*alfa

                    terr = 1. + 0.5*(terr - 1.)
                    terr = max(terr,1.1)
                    nessais += 1
                    perm0 = popt
                    logger.notice("alfa = ", alfa)
                    logger.notice("terr - 1 = ", terr - 1)

                it +=1
            else :
                #with ExclusiveLock(LOCK_NAME):
                with sqlite.connect(self.__database) as conn:
                    cur = conn.cursor()
                    liste_pm = cur.execute('SELECT permeabilite_x, permeabilite_y, OGC_FID FROM mailles').fetchall()
                self.__code.create_permeabilities_sat(sat_dir, liste_pm)
                # calcul hydro
                self.__code.compute_hynverse(sat_dir)

                # calcul des potentiels aux noeuds
                logger.notice(u'Calcul du potentiel aux points pilote')
                potentiel = self.__code.read_potential_at_nodes(sat_dir, [0])

                it +=1
        # fin de la boucle while
        if not self.perm :
            if erreur < errlim:
                logger.notice("Erreur minimale atteinte")
            if alfa < alfmin:
                logger.notice("Pas minimal atteint")
            if it > niter:
                logger.notice('Nombre maximal iterations atteint')
            if nessais > nessaismax:
                logger.notice('Nombre maximal essais atteint')

            itmax = it - 1

        logger.notice(u'Updating database')

        with sqlite.connect(self.__database) as conn:
            cur = conn.cursor()
            # zones
            if not self.perm:
                for i in sorted(dict_zones.keys()):
                    cur.execute("UPDATE zones SET permeabilite=? WHERE numero=?", (dict_zones[i].perm, i))
            self.__code.update_database(conn, sat_dir)


            conn.commit()


def hynverse(dbname, run=False, perm=False, verbose=False):
    """ Launch MHYNVERSE process

    :param dbname: path to database
    :type dbname: string
    :param run: run flag
    :type run: bool
    :param perm: permeability field provided flag
    :type perm: bool
    :param verbose: verbose flag
    :type verbose: bool

    :return: success state
    :rtype: bool
    """
    settings = Settings(os.path.dirname(os.path.abspath(dbname)))
    debug = int(settings.value("General", "debug")) or verbose
    logger.set_level("debug" if debug else "notice")

    logger.notice('*************************')
    logger.notice('*                       *')
    logger.notice('*    H Y N V E R S E    *')
    logger.notice('*                       *')
    logger.notice('*  F.Renard 13/12/2019  *')
    logger.notice('*************************')

    codeHydro = settings.value("General", "codeHydro")
    codeCommand = settings.value("General", "codeHydroCommand")

    if not codeHydro:
        raise RuntimeError(u"le code de calcul n'est pas configuré")

    logger.notice("\n-------------------------")
    logger.notice("CODE HYDRO = ", codeHydro.upper())
    logger.notice("-------------------------\n")

    if codeHydro == 'metis' and os.path.exists(codeCommand):
        code = Metis(dbname, codeCommand)
    elif codeHydro == 'openfoam':
        code = OpenFoam(dbname)
    else:
        raise NotImplementedError(u"code de calcul non encore implémenté")

    _hynverse = Hynverse(code, perm)

    if run and _hynverse.init_flag:
        _hynverse.calcul()

    logger.notice('*****************************')
    logger.notice('*  F I N   H Y N V E R S E  *')
    logger.notice('*****************************')

# run as script if invoqued as such
if __name__ == "__main__":
    import sys
    import sip
    import codecs
    import re
    from ..database import sqlite
    assert(sip.getapi('QVariant') == 2)

    if len(sys.argv) == 2 and sys.argv[1] in ["-h","--help"]:
        help(sys.modules[__name__])
        exit(0)

    logger.enable_console(True)

    if len(sys.argv) not in [2, 3, 4, 5]:
        logger.error("wrong number of arguments (try compute.py --help)")
        exit(1)

    run = False
    verbose = False
    perm=False
    dbname = None
    for arg in sys.argv[1:-1]:
        if arg == '-r': run = True
        elif arg == '-v': verbose = True
        elif arg == '-perm': perm = True

    dbname = sys.argv[-1]

    if not dbname:
        logger.error("cannot find a '.sqlite' file in arguments (try mesh_to_spatialite.py --help)")
        exit(1)

    if False:
        # fix_print_with_import
        print("creating in mem db")
        memdb = "workdb:memory:"
        source_connection = sqlite.connect(dbname)
        dest_conn = sqlite.connect(memdb)
        for line in source_connection.iterdump():
            try:
                if re.match(".*idx_.*_GEOMETRY_.*", line) :
                    #print "HERE", line.encode("utf-8")
                    pass
                else:
                    dest_conn.execute('%s\n' % line)
            except Exception as e:
                # fix_print_with_import
                print(line, "exc", e)
        source_connection.close()

        #script = ''.join(source_connection.iterdump())
        #codecs.open("/tmp/debug", 'w', 'utf-8').write(script)
        #print "dumped"
        #dest_conn.executescript(script)
        #dest_conn.commit()
        #dest_conn.close()
        # fix_print_with_import
        print("done")
        dbname = memdb
        # fix_print_with_import
        print(dbname)


    hynverse(dbname, run, perm, verbose)
